; log.scm
; $Id: log.scm,v 1.12 2005/10/04 14:56:26 lioka Exp $

;------------------------------------------------------------------------------
; transaction logging 
;------------------------------------------------------------------------------
; LOG, pointed by *packages-log-top* is a list of transactions
; TRANSACTION is a cons (command . delta), where
; COMMAND is a list of strings: (cmd . contents)
;         CMD ::= "install" | "remove"
; DELTA is a list of elements like COMMAND, that is,
;       list of lists of strings and represents actual changes
;       between previous and current packages state
; transaction example:
; (("install" "openssh-clients") (("install" "openssh" "openssh-clients")))

(define *packages-log-top* '())
(define *packages-log-point* *packages-log-top*)

(define (packages-log-pending)
  (map
   (lambda (e)
     (let ((address (string-append "/packages/pending/" e)))
       (cons e (sort (woo-read-names address (woo-list address)) string<))))
   (filter
    (lambda (e)
      (member e '("install" "remove" "upgrade")))
    (woo-read-names "/packages/pending"
		    (woo-list "/packages/pending")))))

(define *packages-log-pending* (packages-log-pending))

(define (packages-log-revert-cmd cmd)
  (cond ((string=? cmd "install") "remove")
        ((string=? cmd "remove") "install")
        (else cmd)))

(define (packages-log-revert-command command)
  (cons (packages-log-revert-cmd (car command)) (cdr command)))

(define (packages-log-apply-command command)
  (let ((cmd (car command))
        (contents (cdr command)))
    (map
     (lambda (p)
       (woo-write
        (string-append "/packages/state/" p)
        'state cmd))
     contents)))

(define (packages-log-delta old new)
  (delq #f
        (map
         (lambda (cmd)
           (let* ((old (or (assoc-ref old cmd) '()))
                  (new (or (assoc-ref new cmd) '()))
                  (forward (lset-difference equal? new old))
                  (backward (lset-difference equal? old new)))
             (or
              (and (not (null? forward))
                   (cons (packages-log-revert-cmd cmd) forward))
              (and (not (null? backward))
                   (cons cmd backward)))))
         (lset-union equal?
                     (map car old)
                     (map car new)))))

(define (packages-log-point-up top point)
  (or (and (eq? (cdr top) point) top)
      (packages-log-point-up (cdr top) point)))

;;; reduces log by eliminating complementary commands
(define (packages-log-reduce top)
  (reverse
   (delete-duplicates
    top
    (lambda (x y)
      (let ((x-command (car x))
	    (y-command (packages-log-revert-command (car y))))
	(equal? x-command y-command))))))

(define (packages-log-find top command)
  (or (and (or (null? top)
               (list= equal? command (caar top)))
           top)
      (packages-log-find (cdr top) command)))

;------------------------------------------------------------------------------

(define (packages-log-add command)
  (let* ((pending (packages-log-pending))
	 (delta (packages-log-delta *packages-log-pending* pending)))
    (set! *packages-log-pending* pending)
    (set! *packages-log-point* (cons (cons command delta) *packages-log-point*))
    (set! *packages-log-top* *packages-log-point*)))

(define (packages-log-undo)
  (or (packages-log-undo-empty?)
      (begin
	(map packages-log-apply-command
             (cdar *packages-log-point*))
	(set! *packages-log-point* (cdr *packages-log-point*))
	(set! *packages-log-pending* (packages-log-pending)))))

;;; reduces log, then finds complementary command;
;;; if found, replays log, excluding complementary command
;;; oterwise noop
(define (packages-log-undo-trans command)
  (let* ((rcommand (packages-log-revert-command command))
	 (top (packages-log-reduce *packages-log-point*))
	 (point (packages-log-find top rcommand)))
    (if (null? point) #f
	(let ((rcommand (caar point)))
	  (woo-write "/packages/control" 'control "quit")
	  (map
	   (lambda (command)
	     (or (eq? rcommand command)
		 (packages-log-apply-command command)))
	   (map car top))))))

(define (packages-log-redo)
  (or (packages-log-redo-empty?)
      (begin
	(set! *packages-log-point*
	      (packages-log-point-up *packages-log-top* *packages-log-point*))
        (packages-log-apply-command (caar *packages-log-point*))
        (set! *packages-log-pending* (packages-log-pending)))))

(define (packages-log-set top)
  (set! *packages-log-top* top)
  (set! *packages-log-point* *packages-log-top*)
  (set! *packages-log-pending* (packages-log-pending)))

(define (packages-log-get)
  *packages-log-top*)

(define (packages-log-reset)
  (set! *packages-log-top* '())
  (set! *packages-log-point* *packages-log-top*)
  (set! *packages-log-pending* (packages-log-pending)))

(define (packages-log-undo-empty?)
  (null? *packages-log-point*))

(define (packages-log-redo-empty?)
  (or (null? *packages-log-top*)
      (eq? *packages-log-point* *packages-log-top*)))

(define (packages-log-pending-empty?)
  (null? *packages-log-pending*))
